import * as React from 'react';
import { v4 as uuidv4 } from 'uuid';
import { sendErrorToNoibu } from './core';

// TODO: upload to npm repo in pipeline, maybe put under public_sdks in fe?

export const UNKNOWN_COMPONENT = 'unknown';

// if we pass a non react component we need to make
// sure we pass all the correct information in a function
// which also needs needs to return a react element
export type FallbackRender = (errorData: {
  error: Error;
  componentStack: string | null;
  eventId: string | null;
  resetError(): void;
}) => React.ReactElement;

// props necessary for error boundary element
export type ErrorBoundaryProps = {
  children?: React.ReactNode | (() => React.ReactNode);
  /**
   * A fallback component that gets rendered when the error boundary encounters an error.
   *
   * Can either provide a React Component, or a function that returns React Component as
   * a valid fallback prop. If a function is provided, the function will be called with
   * the error, the component stack, and a function that resets the error boundary on error.
   *
   */
  fallback?: React.ReactElement | FallbackRender;
  /** Called when the error boundary encounters an error */
  onError?(error: Error, componentStack: string, eventId: string): void;
  /** Called on componentDidMount() */
  onMount?(): void;
  /** Called if resetError() is called from the fallback render props function  */
  onReset?(
    error: Error | null,
    componentStack: string | null,
    eventId: string | null,
  ): void;
  /** Called on componentWillUnmount() */
  onUnmount?(
    error: Error | null,
    componentStack: string | null,
    eventId: string | null,
  ): void;
};

// the state type of the error boundary object
type ErrorBoundaryState = {
  componentStack: React.ErrorInfo['componentStack'] | null;
  error: Error | null;
  eventId: string | null;
};

// initial state variables
const INITIAL_STATE = {
  componentStack: null,
  error: null,
  eventId: null,
};

/**
 * A ErrorBoundary component that logs errors to Noibu. Requires React >= 16.
 */
class ErrorBoundary extends React.Component<
  ErrorBoundaryProps,
  ErrorBoundaryState
> {
  public state: ErrorBoundaryState = INITIAL_STATE;

  /**
   * Lifecycle hook on mount
   * @returns void
   */
  public componentDidMount(): void {
    const { onMount } = this.props;
    if (onMount) {
      onMount();
    }
  }

  /**
   *
   * Handler for all errors that happen in a wrapped component
   * @param  {Error&{cause?:Error}} error
   * @param  {React.ErrorInfo} {componentStack}
   * @returns void
   */
  public componentDidCatch(
    error: Error & { cause?: Error },
    { componentStack }: React.ErrorInfo,
  ): void {
    const { onError } = this.props;

    sendErrorToNoibu(error);
    const eventId = uuidv4();
    if (onError) {
      onError(error, componentStack, eventId);
    }

    // componentDidCatch is used over getDerivedStateFromError
    // so that componentStack is accessible through state.
    this.setState({ error, componentStack, eventId });
  }

  /**
   *  * Lifecycle hook on unmount
   * @returns void
   */
  public componentWillUnmount(): void {
    const { error, componentStack, eventId } = this.state;
    const { onUnmount } = this.props;
    if (onUnmount) {
      onUnmount(error, componentStack, eventId);
    }
  }

  /**
   * Callback from fallback to reset the error boundary
   * @param  {} =>void=(
   */
  public resetErrorBoundary: () => void = () => {
    const { onReset } = this.props;
    const { error, componentStack, eventId } = this.state;
    if (onReset) {
      onReset(error, componentStack, eventId);
    }
    this.setState(INITIAL_STATE);
  };

  /**
   *
   * Renders the fallback ui
   * @returns React
   */
  public render(): React.ReactNode {
    const { fallback, children } = this.props;
    const { error, componentStack, eventId } = this.state;

    if (error) {
      let element: React.ReactElement | undefined;
      if (typeof fallback === 'function') {
        element = fallback({
          error,
          componentStack,
          resetError: this.resetErrorBoundary,
          eventId,
        });
      } else {
        element = fallback;
      }

      if (React.isValidElement(element)) {
        return element;
      }

      // Fail gracefully if no fallback provided or is not valid
      return null;
    }

    if (typeof children === 'function') {
      return (children as () => React.ReactNode)();
    }
    return children;
  }
}

export { ErrorBoundary };
